CDORemap <- function(data_array = NULL, lons = NULL, lats = NULL, 
                      grid, method) {  #, mask = NULL) {
  # Check parameters data_array, lons and lats.
  known_lon_names <- .KnownLonNames()
  known_lat_names <- .KnownLatNames()
  if (!is.numeric(lons) || !is.numeric(lats)) {
    stop("Expected numeric 'lons' and 'lats'.")
  }
  if (any(is.na(lons > 0))) {
    stop("Found invalid values in 'lons'.")
  }
  if (any(is.na(lats > 0))) {
    stop("Found invalid values in 'lats'.")
  }
  if (is.null(dim(lons))) {
    dim(lons) <- length(lons)
  }
  if (is.null(dim(lats))) {
    dim(lats) <- length(lats)
  }
  if (length(dim(lons)) > 2 || length(dim(lats)) > 2) {
    stop("'lons' and 'lats' can only have up to 2 dimensions.")
  }
  if (length(dim(lons)) != length(dim(lats))) {
    stop("'lons' and 'lats' must have the same number of dimensions.")
  }
  if (length(dim(lons)) == 2 && !all(dim(lons) == dim(lats))) {
    stop("'lons' and 'lats' must have the same dimension sizes.")
  }
  if (is.null(attr(lons, 'name'))) {
    attr(lons, 'name') <- 'lon'
  }
  if (is.null(attr(lats, 'name'))) {
    attr(lats, 'name') <- 'lat'
  }
  return_array <- TRUE
  if (is.null(data_array)) {
    return_array <- FALSE
    if (length(dim(lons)) == 1) {
      array_dims <- c(length(lats), length(lons))
    } else {
      array_dims <- dim(lons)
    }
    data_array <- array(NA, array_dims)
    names(dim(data_array)) <- c('lat', 'lon')
  }
  if (!(is.logical(data_array) || is.numeric(data_array)) || !is.array(data_array)) {
    stop("Parameter 'data_array' must be a numeric array.")
  }
  if (is.null(names(dim(data_array)))) {
    stop("Parameter 'data_array' must have named dimensions.")
  }
  # Cannot do this! data_array would be duplicated.
  #if (is.null(attr(data_array, 'name'))) {
  #  attr(data_array, 'name') <- 'unnamed_var'
  #}
  var_name <- 'unnamed_var'
  if (!is.null(attr(data_array, 'name'))) {
    var_name <- attr(data_array, 'name')
  }
  lon_dim <- which(known_lon_names %in% names(dim(data_array)))
  if (length(lon_dim) < 1) {
    stop("Could not find a known longitude dimension name in the provided 'data_array'.")
  }
  if (length(lon_dim) > 1) {
    stop("Found more than one known longitude dimension names in the provided 'data_array'.")
  }
  lon_dim <- known_lon_names[lon_dim]
  lat_dim <- which(known_lat_names %in% names(dim(data_array)))
  if (length(lat_dim) < 1) {
    stop("Could not find a known latitude dimension name in the provided 'data_array'.")
  }
  if (length(lat_dim) > 1) {
    stop("Found more than one known latitude dimension name in the provided 'data_array'.")
  }
  lat_dim <- known_lat_names[lat_dim]
  # Check grid.
  if (grepl('^r[0-9]{1,}x[0-9]{1,}$', grid)) {
    grid_type <- 'regular'
    grid_lons <- strsplit(strsplit(grid, 'x')[[1]][1], 'r')[[1]][2]
    grid_lats <- strsplit(grid, 'x')[[1]][2]
  } else if (grepl('^t[0-9]{1,}grid$', grid)) {
    grid_type <- 'gaussian'
    grid_t <- strsplit(strsplit(grid, 'grid')[[1]][1], 't')[[1]][2]
    grid_size <- .t2nlatlon(grid_t)
    grid_lons <- grid_size[2]
    grid_lats <- grid_size[1]
  } else {
    grid_type <- 'custom'
  }
  # Check method.
  if (method %in% c('bil', 'bilinear')) {
    method <- 'bil'
  } else if (method %in% c('bic', 'bicubic')) {
    method <- 'bic'
  } else if (method %in% c('con', 'conservative')) {
    method <- 'con'
  } else if (method %in% c('dis', 'distance-weighted')) {
    method <- 'dis'
  } else {
    stop("Unsupported CDO remap method. 'bilinear', 'bicubic', 'conservative' or 'distance-weighted' supported only.")
  }
#  if (!is.null(mask)) {
#    if (!is.numeric(mask) || !is.array(mask)) {
#      stop("Parameter 'mask' must be a numeric array.")
#    }
#    if (length(dim(mask)) != 2) {
#      stop("Parameter 'mask' must have two dimensions.")
#    }
#    if (is.null(names(dim(mask)))) {
#      if (dim(data_array)[lat_dim] == dim(data_array)[lon_dim]) {
#        stop("Cannot disambiguate which is the longitude dimension of ",
#             "the provided 'mask'. Provide it with dimension names.")
#      }
#      names(dim(mask)) <- c('', '')
#      found_lon_dim <- which(dim(mask) == dim(data_array)[lon_dim])
#      if (length(found_lon_dim) < 0) {
#        stop("The dimension sizes of the provided 'mask' do not match ",
#             "the spatial dimension sizes of the array to interpolate.")
#      } else {
#        names(dim(mask)[found_lon_dim]) <- lon_dim
#      }
#      found_lat_dim <- which(dim(mask) == dim(data_array)[lat_dim])
#      if (length(found_lat_dim) < 0) {
#        stop("The dimension sizes of the provided 'mask' do not match ",
#             "the spatial dimension sizes of the array to interpolate.")
#      } else {
#        names(dim(mask)[found_lat_dim]) <- lat_dim
#      }
#    }
#    lon_position <- which(names(dim(data_array)) == lon_dim)
#    lat_position <- which(names(dim(data_array)) == lat_dim)
#    if (lon_position > lat_position) {
#      if (names(dim(mask))[1] == lon_dim) {
#        mask <- t(mask)
#      }
#    } else {
#      if (names(dim(mask))[1] == lat_dim) {
#        mask <- t(mask)
#      }
#    }
#    ## TODO: Apply mask!!! Preserve attributes
#  }
  # Check if interpolation can be skipped.
  interpolation_needed <- TRUE
  if (length(lons) == grid_lons && length(lats) == grid_lats) {
    lon_spaces <- lons[2:length(lons)] - lons[1:(length(lons) - 1)]
    lat_spaces <- lats[2:length(lats)] - lats[1:(length(lats) - 1)]
    if (grid_type == 'regular') {
      if ((count(abs(lon_spaces - mean(lon_spaces)) > mean(lon_spaces) / 10) < 2) &&
          (count(abs(lat_spaces - mean(lat_spaces)) > mean(lat_spaces) / 10) < 1)) {
        interpolation_needed <- FALSE
      }
    } else if (grid_type == 'gaussian') {
      # TODO: improve this check. Gaussian quadrature should be used.
      if ((count(abs(lon_spaces - mean(lon_spaces)) > mean(lon_spaces) / 10) < 2) &&
          (count(abs(lat_spaces - mean(lat_spaces)) > mean(lat_spaces) / 10) > 0)) {
        interpolation_needed <- FALSE
      }
    }
  }
  if (interpolation_needed) {
    # CDO takes arrays of 3 dimensions or 4 if one of them is unlimited.
    # The unlimited dimension can only be the left-most (right-most in R).
    # There are no restrictions for the dimension names or variable names.
    # The longitude and latitude are detected by their units.
    # There are no restrictions for the order of the limited dimensions.
    # The procedure consists in:
    # - take out the array metadata
    # - be aware of var dimension (replacing the dimension names would do).
    # - take arrays of 4 dimensions always if possible
    # - make the last dimension unlimited when saving to netcdf
    # - if the last dimension is lon or lat, either reorder the array and 
    #   then reorder back or iterate over the dimensions at the right
    #   side of lon AND lat.
    # If the input array has more than 4 dimensions, it is needed to
    # run CDO on each sub-array of 4 dimensions because it can handle
    # only up to 4 dimensions. The shortest dimensions are chosen to 
    # iterate over.
    other_dims <- which(!(names(dim(data_array)) %in% c(lon_dim, lat_dim)))
    dims_to_iterate <- NULL
    if (length(other_dims) > 2) {
      dims_to_iterate <- other_dims[tail(sort(dim(data_array)[other_dims], index.return = TRUE)$ix, 2)]
      slices_to_iterate <- array(1:prod(dim(data_array)[dims_to_iterate]), 
                                  dim(data_array)[dims_to_iterate])
      total_slices <- prod(dim(slices_to_iterate))
    } else {
      total_slices <- 1
    }
    result_array <- NULL
    for (i in 1:total_slices) {
      tmp_file <- tempfile('R_CDORemap_', fileext = '.RData')
      tmp_file2 <- tempfile('R_CDORemap_', fileext = '.RData')
      if (!is.null(dims_to_iterate)) {
        slice_indices <- which(slices_to_iterate == i, arr.ind = TRUE)
        subset <- Subset(data_array, dims_to_iterate, as.list(slice_indices))
        ArrayToNetCDF(list(subset, lons, lats), tmp_file)
      } else {
        ArrayToNetCDF(list(data_array, lons, lats), tmp_file)
      }
      if (nchar(Sys.which('cdo')[1]) < 1) {
        stop("CDO must be installed to use the .CDORemap.")
      }
      system(paste0("cdo -s remap", method, ",", grid, " ", tmp_file, " ", tmp_file2, " 2> /dev/null"))
      file.remove(tmp_file)
      ncdf_remapped <- nc_open(tmp_file2)
      if (!is.null(dims_to_iterate)) {
        if (is.null(result_array)) {
          found_lons <- ncdf_remapped$dim$lon$vals
          found_lats <- ncdf_remapped$dim$lat$vals
          if (return_array) {
            new_dims <- dim(data_array)
            new_dims[c(lon_dim, lat_dim)] <- c(ncdf_remapped$dim$lon$len, ncdf_remapped$dim$lat$len)
            result_array <- array(dim = new_dims)
            store_indices <- as.list(rep(TRUE, length(dim(result_array))))
          }
        }
        if (return_array) {
          store_indices[dims_to_iterate] <- as.list(slice_indices)
          result_array <- do.call('[<-', c(list(x = result_array), store_indices, 
                                           list(value = ncvar_get(ncdf_remapped, var_name))))
        }
      } else {
        found_lons <- ncdf_remapped$dim$lon$vals
        found_lats <- ncdf_remapped$dim$lat$vals
        if (return_array) {
          result_array <- ncvar_get(ncdf_remapped, var_name)
        }
      }
      nc_close(ncdf_remapped)
    }
  }
  #    what about inputting sparse data? a square will be generated? is it ok?
  list(data_array = ifelse(return_array, return_array, NULL), 
       lons = found_lons, lats = found_lats)
}
